//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "LastCommonRouteFinder.h"
#include "Pinger.h"
#include "EMule.h"


LastCommonRouteFinder::LastCommonRouteFinder() {
    minUpload = 1;
    maxUpload = _UI16_MAX;
    m_upload = _UI16_MAX;
    m_CurUpload = 1;

    m_pingAverage = 0;
    pingDelaysTotal = 0;

    needMoreHosts = false;

    threadEndedEvent = new CEvent(0, 1);
    newTraceRouteHostEvent = new CEvent(0, 0);
    prefsEvent = new CEvent(0, 0);

    m_enabled = false;
    doRun = true;
    AfxBeginThread(RunProc, (LPVOID)this);
}

LastCommonRouteFinder::~LastCommonRouteFinder() {
    delete threadEndedEvent;
    delete newTraceRouteHostEvent;
    delete prefsEvent;
}

bool LastCommonRouteFinder::AddHostsToCheck(CTypedPtrList<CPtrList, CServer*> &list) {
    if(list.GetCount() >= 10 && needMoreHosts) {
        addHostLocker.Lock();
        
        if(needMoreHosts) {
            hostsToTraceRoute.RemoveAll();

            uint32 startPos = rand()/(RAND_MAX/list.GetCount());

            POSITION pos = list.GetHeadPosition();
            for(uint32 skipCounter = startPos; skipCounter < (uint32)list.GetCount() && pos != NULL; skipCounter++) {
                list.GetNext(pos);
            }

            uint32 tryCount = 0;
            while(pos != NULL && hostsToTraceRoute.GetCount() < 10 && tryCount <= (uint32)list.GetCount()) {
                tryCount++;
                CServer* server = list.GetNext(pos);

                uint32 ip = server->GetIP();

                if(IsGoodIP(ip, true)) {
                    hostsToTraceRoute.AddTail(ip);
                }

                if(pos == NULL) {
                    POSITION pos = list.GetHeadPosition();
                }
            }

            if(hostsToTraceRoute.GetCount() >= 10) {
                needMoreHosts = false;

                // Signal that there's hosts to fetch.
                newTraceRouteHostEvent->SetEvent();

                addHostLocker.Unlock();
                return true;
            }
        }

        addHostLocker.Unlock();
    }

    return false;
}

bool LastCommonRouteFinder::AddHostsToCheck(CTypedPtrList<CPtrList, CUpDownClient*> &list) {
    if(list.GetCount() >= 10 && needMoreHosts) {
        addHostLocker.Lock();
        
        if(needMoreHosts) {
            hostsToTraceRoute.RemoveAll();

            uint32 startPos = rand()/(RAND_MAX/list.GetCount());

            POSITION pos = list.GetHeadPosition();
            for(uint32 skipCounter = startPos; skipCounter < (uint32)list.GetCount() && pos != NULL; skipCounter++) {
                list.GetNext(pos);
            }

            uint32 tryCount = 0;
            while(pos != NULL && hostsToTraceRoute.GetCount() < 10 && tryCount <= (uint32)list.GetCount()) {
                tryCount++;
                CUpDownClient* client = list.GetNext(pos);

                uint32 ip = client->GetIP();

                if(IsGoodIP(ip, true)) {
                    hostsToTraceRoute.AddTail(ip);
                }

                if(pos == NULL) {
                    POSITION pos = list.GetHeadPosition();
                }
            }

            if(hostsToTraceRoute.GetCount() >= 10) {
                needMoreHosts = false;

                // Signal that there's hosts to fetch.
                newTraceRouteHostEvent->SetEvent();
                return true;
            }
        }

        addHostLocker.Unlock();
    }
    return false;
}


uint32 LastCommonRouteFinder::GetCurrentPing() {
    uint32 returnVal = 0;

    if(m_enabled) {
        pingLocker.Lock();
        returnVal = m_pingAverage;
        pingLocker.Unlock();
    }

    return returnVal;
}

uint32 LastCommonRouteFinder::GetPingedHost() {
    return 0;
}

bool LastCommonRouteFinder::AcceptNewClient() {
    return acceptNewClient || !m_enabled; // if enabled, then return acceptNewClient, otherwise return true
}

void LastCommonRouteFinder::SetPrefs(bool pEnabled, uint16 pCurUpload, uint16 pMinUpload, uint16 pMaxUpload, double pPingTolerance, int pGoingUpDivider, int pGoingDownDivider) {
    bool sendEvent = false;

    prefsLocker.Lock();

    if(pMinUpload >= 1) {
        minUpload = 1;
    } else {
        minUpload = pMinUpload;
    }

    if(pMaxUpload != 0) {
        maxUpload = pMaxUpload;
        if(maxUpload < minUpload) {
            minUpload = maxUpload;
        }
    } else {
        maxUpload = _UI16_MAX;
    }

    if(pEnabled && m_enabled == false) {
        sendEvent = true;
    } else if(pEnabled == false) {
        prefsEvent->ResetEvent();
    }
    m_enabled = pEnabled;
    m_pingTolerance = pPingTolerance;
    m_goingUpDivider = pGoingUpDivider;
    m_goingDownDivider = pGoingDownDivider;
    m_CurUpload = pCurUpload;

    uploadLocker.Lock();

    if (m_upload > maxUpload || pEnabled == false) {
        m_upload = maxUpload;
    }

    uploadLocker.Unlock();
    prefsLocker.Unlock();

    if(sendEvent) {
        prefsEvent->SetEvent();
    }
}

uint16 LastCommonRouteFinder::GetUpload() {
    uint16 returnValue;

    uploadLocker.Lock();

    returnValue = m_upload;

    uploadLocker.Unlock();

    return returnValue;
}

void LastCommonRouteFinder::SetUpload(uint16 newValue) {
    uploadLocker.Lock();

    m_upload = newValue;

    uploadLocker.Unlock();
}

/**
 * Make the thread exit. This method will not return until the thread has stopped
 * looping.
 */
void LastCommonRouteFinder::EndThread() {
    // signal the thread to stop looping and exit.
    doRun = false;

    prefsEvent->SetEvent();
    newTraceRouteHostEvent->SetEvent();

    // wait for the thread to signal that it has stopped looping.
    threadEndedEvent->Lock();
}

/**
 * Start the thread. Called from the constructor in this class.
 *
 * @param pParam
 *
 * @return
 */
UINT AFX_CDECL LastCommonRouteFinder::RunProc(LPVOID pParam) {
    LastCommonRouteFinder* lastCommonRouteFinder = (LastCommonRouteFinder*)pParam;

    return lastCommonRouteFinder->RunInternal();
}

/**
 * @return always returns 0.
 */
UINT LastCommonRouteFinder::RunInternal() {
    Pinger pinger;

    while(doRun) {
        bool foundLastCommonHost = false;
        uint32 lastCommonHost = 0;

        // wait for updated prefs
        prefsEvent->Lock();

        bool enabled = m_enabled;

        hostsToTraceRoute.RemoveAll();

        pingDelays.RemoveAll();
        pingDelaysTotal = 0;

        pingLocker.Lock();
        m_pingAverage = 0;
        pingLocker.Unlock();

        int traceRouteTries = 0;
        while(doRun && enabled && foundLastCommonHost == false && traceRouteTries < 3) {
            lastCommonHost = 0;

            theApp.emuledlg->QueueDebugLogLine(false,"Traceroute: Collecting hosts...");

            addHostLocker.Lock();
            needMoreHosts = true;
            addHostLocker.Unlock();

            // wait for hosts to traceroute
            newTraceRouteHostEvent->Lock();

            theApp.emuledlg->QueueDebugLogLine(false,"Traceroute: Got enough hosts. Listing the hosts that will be tracerouted:");

            {
                POSITION pos = hostsToTraceRoute.GetHeadPosition();
                int counter = 0;
                while(pos != NULL) {
                    counter++;
                    uint32 hostToTraceRoute = hostsToTraceRoute.GetNext(pos);
                    IN_ADDR stDestAddr;
                    stDestAddr.s_addr = hostToTraceRoute;

                    theApp.emuledlg->QueueDebugLogLine(false,"Host #%i: %s", counter, inet_ntoa(stDestAddr));
                }
            }

            // find the last common host, using traceroute

            theApp.emuledlg->QueueDebugLogLine(false,"Traceroute: Starting traceroutes to find last common host.");

            if(m_enabled == false) {
                enabled = false;
            }

            bool failed = false;

            for(uint32 ttl = 1; doRun && enabled && ttl <= 64 && foundLastCommonHost == false && failed == false; ttl++) {
                uint32 curHost = 0;
                PingStatus pingStatus;

                if(m_enabled == false) {
                    enabled = false;
                }

                bool firstLoop = true;
                POSITION pos = hostsToTraceRoute.GetHeadPosition();
                while(doRun && enabled && failed == false && pos != NULL &&
                      (
                       firstLoop ||
                       pingStatus.success && pingStatus.destinationAddress == curHost
                      )
                     ) {

                    firstLoop = false;

                    POSITION lastPos = pos;
                    uint32 curAddress = hostsToTraceRoute.GetNext(pos);

                    pingStatus.success = false;
                    for(int counter = 0; doRun && enabled && counter < 3 && pingStatus.success == false; counter++) {
                        pingStatus = pinger.Ping(curAddress, ttl, true);
                        if(doRun && enabled && pingStatus.success == false && counter < 3-1) {
                            IN_ADDR stDestAddr;
                            stDestAddr.s_addr = curAddress;
                            theApp.emuledlg->QueueDebugLogLine(false,"Traceroute: Failure #%i to ping host! (TTL: %i IP: %s error: %i). Sleeping 1 sec before retry.", counter+1, ttl, inet_ntoa(stDestAddr), pingStatus.error);
                            Sleep(1000);

                            if(m_enabled == false) {
                                enabled = false;
                            }
                        }
                    }

                    if(pingStatus.success == true && pingStatus.status != IP_SUCCESS) {
                        if(curHost == 0) {
                            curHost = pingStatus.destinationAddress;
                        }
                    } else {
                        // failed to ping this host for some reason.
                        // Or we reached the actual host we are pinging. We don't want that, since it is too close.
                        // Remove it.
                        hostsToTraceRoute.RemoveAt(lastPos);
                        failed = true;

                        IN_ADDR stDestAddr;
                        stDestAddr.s_addr = curAddress;
                        if(pingStatus.success == true) {
                            theApp.emuledlg->QueueDebugLogLine(false,"Traceroute: Host was too close! (TTL: %i IP: %s status: %i). Removing this host and restarting host collection.", ttl, inet_ntoa(stDestAddr), pingStatus.status);
                        } else {
                            theApp.emuledlg->QueueDebugLogLine(false,"Traceroute: Error when pinging a host! (TTL: %i IP: %s Error: %i). Removing this host and restarting host collection.", ttl, inet_ntoa(stDestAddr), pingStatus.error);
                        }
                    }
                }

                if(failed == false) {
                    if(pingStatus.success && pingStatus.destinationAddress == curHost) {
                        IN_ADDR stDestAddr;
                        stDestAddr.s_addr = curHost;
                        theApp.emuledlg->QueueDebugLogLine(false,"Traceroute: Host at TTL %i: %s", ttl, inet_ntoa(stDestAddr));

                        lastCommonHost = curHost;
                    } else {
                        foundLastCommonHost = true;
                    }
                }
            }
        }

        if(foundLastCommonHost == true) {
            IN_ADDR stDestAddr;
            stDestAddr.s_addr = lastCommonHost;

            // log result
            theApp.emuledlg->QueueDebugLogLine(false,"Traceroute: Found last common host. IP: %s", inet_ntoa(stDestAddr));
        } else {
            theApp.emuledlg->QueueDebugLogLine(false,"Traceroute: Failed to many times. Disabling Upload SpeedSense.");
            enabled = false;

            // PENDING: this may not be thread safe
            m_enabled = false;
        }

        if(m_enabled == false) {
            enabled = false;
        }


        uint32 initial_ping = 9999;

        uploadLocker.Lock();
        for(int initialPingCounter = 0; doRun && enabled && initialPingCounter < 10; initialPingCounter++) {
            Sleep(200);

            // finding lowest ping
            PingStatus pingStatus = pinger.Ping(lastCommonHost);

            if (pingStatus.success && initial_ping > pingStatus.delay) {
    		    initial_ping = pingStatus.delay;
		    }

            if(m_enabled == false) {
                enabled = false;
            }
        }
        uploadLocker.Unlock();

        double upload;

        prefsLocker.Lock();
        upload = m_CurUpload;

        if(upload > maxUpload) {
            upload = maxUpload;
        }
        
        if(upload < 1) {
            upload = 1;
        }
        prefsLocker.Unlock();

        if(m_enabled == false) {
            enabled = false;
        }

        while(doRun && enabled) {
            if(upload <= 2) {
                Sleep(500);
            } else if(upload >= 20) {
                Sleep(250);
            } else if(upload >= 50) {
                Sleep(125);
            }

            prefsLocker.Lock();
            int pingTolerance = m_pingTolerance;
            double goingUpDivider = m_goingUpDivider;
            double goingDownDivider = m_goingDownDivider;
            prefsLocker.Unlock();

            int soll_ping = initial_ping*pingTolerance;

            // ping the found last common host
            PingStatus pingStatus = pinger.Ping(lastCommonHost);

            uint32 raw_ping;
                
            if(pingStatus.success) {
                raw_ping = (uint32)pingStatus.delay;
            } else {
                raw_ping = soll_ping*3;
            }

		    if (initial_ping > raw_ping) {
			    initial_ping = raw_ping;
		    }

            pingDelaysTotal += raw_ping;
            pingDelays.AddTail(raw_ping);
            while(pingDelays.GetCount() > 4) {
                uint32 pingDelay = pingDelays.RemoveHead();
                pingDelaysTotal -= pingDelay;
            }

		    //int normalized_ping = raw_ping - initial_ping;
		    int normalized_ping = (pingDelaysTotal/pingDelays.GetCount()) - initial_ping;

            pingLocker.Lock();
            m_pingAverage = pingDelaysTotal/pingDelays.GetCount();
            pingLocker.Unlock();

		    // Calculate Waiting Time
		    int hping = (soll_ping - normalized_ping);
    		
            double ulDiff;
            if(hping < 0) {
                // lower the speed
                ulDiff = (double)hping / goingDownDivider;
                acceptNewClient = false;
            } else {
                // raise the speed
                ulDiff = (double)hping / goingUpDivider;
                acceptNewClient = true;
            }

		    upload += ulDiff;

		    if (upload < 1) {
			    upload = 1;
            } else {
                prefsLocker.Lock();

                if (upload > maxUpload) {
			        upload = maxUpload;
                }

                prefsLocker.Unlock();
            }

            SetUpload((uint16)upload);

            if(m_enabled == false) {
                enabled = false;
            }
        }
    }

    // Signal that we have ended.
    threadEndedEvent->SetEvent();

    return 0;
}